import pandas as pd

from .client import BaseClient


class SummaryClient(BaseClient):
    """
    Client class for summary requests.
    """

    def create_visual(self, data: dict) -> dict:
        """Create a new visual

        Args:
            data (dict): Required metadata for object creation.
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - visual_type (int): Constant representing visual type
                - evidence_type (int): Constant representing evidence type
                - published (bool): visual is published for public view
                - settings (dict): object settings (must be valid JSON)
                - assessment (int): assessment ID
                - prefilters (dict): object prefilters (must be valid JSON)
                - caption (str): Visual caption

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/summary/api/visual/"
        return self.session.post(url, data=data).json()

    def update_visual(self, visual_id: int, data: dict) -> dict:
        """Create a new visual

        Args:
            id (int): Visual identifier
            data (dict): Metadata to update
                - title (str): Visual title
                - slug (str): Visual identifier/URL base
                - visual_type (int): Constant representing visual type
                - evidence_type (int): Constant representing evidence type
                - published (bool): visual is published for public view
                - settings (dict): object settings (must be valid JSON)
                - assessment (int): assessment ID
                - prefilters (dict): object prefilters (must be valid JSON)
                - caption (str): Visual caption

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        return self.session.patch(url, data=data).json()

    def delete_visual(self, visual_id: int):
        """Delete a visual.

        Args:
            visual_id (int): ID of the visual to delete

        Returns:
            None: If the operation is successful there is no return value.
            If the operation is unsuccessful, an error will be raised.
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        self.session.delete(url)

    def get_visual(self, visual_id: int) -> dict:
        """Get a visual.

        Args:
            visual_id (int): ID of the visual to read

        Returns:
            Response: A response object, which contains the visual if successful.
        """
        url = f"{self.session.root_url}/summary/api/visual/{visual_id}/"
        return self.session.get(url).json()

    def visual_list(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves a visual list for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Visual list
        """
        url = f"{self.session.root_url}/summary/api/visual/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def table_list(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves a table list for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Data Pivot list
        """
        url = f"{self.session.root_url}/summary/api/summary-table/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)



import json
import math
from collections.abc import Generator

from requests import Response, Session
from tqdm import tqdm

from .exceptions import HawcClientException, HawcServerException


class HawcSession:
    """
    A session that handles HAWC requests and responses.

    Allows user authentication and keeps track of root url.
    """

    def __init__(self, root_url: str = "https://hawcproject.org"):
        self.root_url = root_url
        self._session = Session()

    def _handle_hawc_response(self, response: Response) -> None:
        """
        Handle standard hawc API response

        Args:
            response (Response): A response object

        Raises:
            HawcClientException: If error occurs in data submission
            HawcServerException: If error orccurs on server
        """
        try:
            content = response.json()
        except json.JSONDecodeError:
            content = response.text
        if response.status_code >= 400 and response.status_code < 500:
            raise HawcClientException(response.status_code, content)
        elif response.status_code >= 500 and response.status_code < 600:
            raise HawcServerException(response.status_code, "no additional information provided")

    def get(self, url: str, params: dict | None = None) -> Response:
        """
        Sends a GET request using the session instance

        Args:
            url (str): URL for request.
            params (dict, optional): Additional parameters to include. Defaults to None.

        Returns:
            Response: The response.
        """
        response = self._session.get(url=url, params=params)
        self._handle_hawc_response(response)
        return response

    def delete(self, url: str, params: dict | None = None) -> Response:
        """
        Sends a DELETE request using the session instance

        Args:
            url (str): URL for request.
            params (dict, optional): Additional parameters to include. Defaults to None.

        Returns:
            Response: The response.
        """
        response = self._session.delete(url=url, params=params)
        self._handle_hawc_response(response)
        return response

    def post(self, url: str, data: dict | None = None) -> Response:
        """
        Sends a POST request using the session instance

        Args:
            url (str): URL for request.
            data (dict, optional): Payload for the request.

        Returns:
            Response: The response.
        """
        response = self._session.post(url=url, json=data)
        self._handle_hawc_response(response)
        return response

    def patch(self, url: str, data: dict | None = None) -> Response:
        """
        Sends a PATCH request using the session instance

        Args:
            url (str): URL for request.
            data (dict, optional): Payload for the request.

        Returns:
            Response: The response.
        """
        response = self._session.patch(url=url, json=data)
        self._handle_hawc_response(response)
        return response

    def authenticate(self, email: str, password: str):
        """
        Authenticate a user session

        Args:
            email (str): email to authenticate
            password (str): password to authenticate
        """
        url = f"{self.root_url}/user/api/token-auth/"
        data = {"username": email, "password": password}
        response = self._session.post(url, json=data)
        self._handle_hawc_response(response)
        token = response.json()["token"]
        self._session.headers.update(Authorization=f"Token {token}")

    def set_authentication_token(self, token: str, login: bool = False) -> bool:
        """
        Set authentication token for hawc client session.

        Args:
            token (str): authentication token from your user profile
            login (bool, default False): if True, creates a django cookie-based session for HAWC
                similar to a standard web-browser. If False (default), creates a token-
                based django session for using the API. Requests using cookie-based sessions
                require CSRF tokens, so form functionality will be limited.

        Returns
            bool: Returns true if session is valid
        """
        params = {}
        if login:
            params["login"] = 1
        self._session.headers.update(Authorization=f"Token {token}")
        url = f"{self.root_url}/user/api/validate-token/"
        response = self.get(url, params).json()
        if login:
            self._session.headers.pop("Authorization")
        return response["valid"]

    def iter_pages(self, url: str, params: dict | None = None) -> Generator:
        """
        Generator that crawls paginated HAWC responses.

        Args:
            url (str): URL for GET request.
            params (dict, optional): GET parameters to include. Defaults to None.

        Returns:
            Generator: Generator for paginated response

        Yields:
            Generator: Results for a page of response
        """
        response_json = self.get(url, params).json()
        yield response_json["results"]
        # Prevents divide by zero if there are no results
        if len(response_json["results"]) == 0:
            return
        num_pages = math.ceil(response_json["count"] / len(response_json["results"]))
        with tqdm(desc="Iterating pages", initial=1, total=num_pages) as progress_bar:
            while response_json["next"] is not None:
                response_json = self.get(response_json["next"]).json()
                progress_bar.update(1)
                yield response_json["results"]



import pandas as pd
from requests import Response

from .client import BaseClient


class EpiClient(BaseClient):
    """
    Client class for epidemiology requests.
    """

    def metadata(self, assessment_id: int) -> dict:
        """
        Retrieves field choices for all epi models.

        Returns:
            dict: Model metadata
        """
        url = f"{self.session.root_url}/epi/api/metadata/{assessment_id}/"
        return self.session.get(url).json()

    def data(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves epidemiology data for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Epidemiology data
        """
        url = f"{self.session.root_url}/epi/api/assessment/{assessment_id}/export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def endpoints(self, assessment_id: int) -> list[dict]:
        """
        Retrieves all of the epidemiology endpoints for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            list[dict]: Epidemiology endpoints
        """
        payload = {"assessment_id": assessment_id}
        url = f"{self.session.root_url}/epi/api/outcome/"
        generator = self.session.iter_pages(url, payload)
        return [res for results in generator for res in results]

    def get_study_populations(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves all of the study populations for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: study population data
        """
        url = f"{self.session.root_url}/epi/api/study-population/?assessment_id={assessment_id}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def get_study_population(self, study_population_id: int) -> dict:
        """
        Retrieves data for a single study population

        Args:
            study_population_id (int): Study Population ID

        Returns:
           dict: study population data
        """
        url = f"{self.session.root_url}/epi/api/study-population/{study_population_id}/"
        return self.session.get(url).json()

    def create_study_population(self, data: dict) -> dict:
        """
        Create a new study population.

        Args:
            data (dict): fields to create on the study population. Example keys:
                name (str): name of the study population
                study (int): id of the study to associate with this study population
                countries (list[str]): list of country codes to associate with the study pop
                design (str): Study Design (CO == Cohort, RT == Randomized controlled trial, etc.)
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/study-population/"
        return self.session.post(url, data).json()

    def update_study_population(self, study_population_id: int, data: dict) -> dict:
        """
        Update an existing study population.

        Args:
            study_population_id (int): Study Population ID

            data: fields to update in the study pop.
                See "create_study_population" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/study-population/{study_population_id}/"
        return self.session.patch(url, data).json()

    def delete_study_population(self, study_population_id: int) -> Response:
        """
        Delete a study population.

        Args:
            study_population_id (int): Study Population ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/study-population/{study_population_id}/"
        return self.session.delete(url)

    def get_criteria(self, criteria_id: int) -> dict:
        """
        Retrieves data for a single criteria

        Args:
            criteria_id (int): criteria ID

        Returns:
            dict: criteria data
        """
        url = f"{self.session.root_url}/epi/api/criteria/{criteria_id}/"
        return self.session.get(url).json()

    def create_criteria(self, data: dict) -> dict:
        """
        Create a new criteria

        Args:
            data (dict): fields to create on the criteria. Keys:
                description (str): description of the criteria
                assessment (int): id of the associated assessment

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/criteria/"
        return self.session.post(url, data).json()

    def update_criteria(self, criteria_id: int, data: dict) -> dict:
        """
        Update an existing criteria

        Args:
            criteria_id (int): criteria ID

            data: fields to update in the criteria
                See "create_criteria" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/criteria/{criteria_id}/"
        return self.session.patch(url, data).json()

    def delete_criteria(self, criteria_id: int) -> Response:
        """
        Delete a criteria

        Args:
            criteria_id (int): criteria ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/criteria/{criteria_id}/"
        return self.session.delete(url)

    def get_comparison_set(self, comparison_set_id: int) -> dict:
        """
        Retrieves data for a single comparison set

        Args:
            comparison_set_id (int): comparison set ID

        Returns:
            dict: comparison set data
        """
        url = f"{self.session.root_url}/epi/api/comparison-set/{comparison_set_id}/"
        return self.session.get(url).json()

    def create_comparison_set(self, data: dict) -> dict:
        """
        Create a new comparison set

        Args:
            data (dict): fields to create on the comparison set. Keys:
                name (str): name of the comparison set
                description (str): description
                exposure (int): id of the associated exposure population
                study_population (int): id of the associated study population

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/comparison-set/"
        return self.session.post(url, data).json()

    def update_comparison_set(self, comparison_set_id: int, data: dict) -> dict:
        """
        Update an existing comparison set

        Args:
            comparison_set_id (int): comparison set ID

            data: fields to update in the comparison set
                See "create_comparison_set" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/comparison-set/{comparison_set_id}/"
        return self.session.patch(url, data).json()

    def delete_comparison_set(self, comparison_set_id: int) -> Response:
        """
        Delete a comparison set

        Args:
            comparison_set_id (int): comparison set ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/comparison-set/{comparison_set_id}/"
        return self.session.delete(url)

    def get_group(self, group_id: int) -> dict:
        """
        Retrieves data for a single group

        Args:
            group_id (int): group ID

        Returns:
            dict: group data
        """
        url = f"{self.session.root_url}/epi/api/group/{group_id}/"
        return self.session.get(url).json()

    def create_group(self, data: dict) -> dict:
        """
        Create a new group

        Args:
            data (dict): fields to create on the group. Example keys:
                name (str): name of the group
                comparison_set (int): id of the associated comparison set
                ethnicities (list[Any]): list of id's/names of associated ethnicities
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/group/"
        return self.session.post(url, data).json()

    def update_group(self, group_id: int, data: dict) -> dict:
        """
        Update an existing group

        Args:
            group_id (int): group ID

            data: fields to update in the group
                See "create_group" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/group/{group_id}/"
        return self.session.patch(url, data).json()

    def delete_group(self, group_id: int) -> Response:
        """
        Delete a group

        Args:
            group_id (int): group ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/group/{group_id}/"
        return self.session.delete(url)

    def get_numerical_description(self, numerical_description_id: int) -> dict:
        """
        Retrieves data for a single numerical_description

        Args:
            numerical_description_id (int): numerical_description ID

        Returns:
            dict: numerical_description data
        """
        url = f"{self.session.root_url}/epi/api/numerical-descriptions/{numerical_description_id}/"
        return self.session.get(url).json()

    def create_numerical_description(self, data: dict) -> dict:
        """
        Create a new numerical description

        Args:
            data (dict): fields to create on the numerical description. Example keys:
                description (str): description
                group (int): id of the associated group
                mean (Number): mean value
                mean (Any): median, geometric mean, etc. readable name or underlying id
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/numerical-descriptions/"
        return self.session.post(url, data).json()

    def update_numerical_description(self, numerical_description_id: int, data: dict) -> dict:
        """
        Update an existing numerical description

        Args:
            numerical_description_id (int): numerical description ID

            data: fields to update in the numerical description
                See "create_numerical_description" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/numerical-descriptions/{numerical_description_id}/"
        return self.session.patch(url, data).json()

    def delete_numerical_description(self, numerical_description_id: int) -> Response:
        """
        Delete a numerical description

        Args:
            numerical_description_id (int): numerical description ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/numerical-descriptions/{numerical_description_id}/"
        return self.session.delete(url)

    def get_outcome(self, outcome_id: int) -> dict:
        """
        Retrieves data for a single outcome

        Args:
            outcome_id (int): outcome ID

        Returns:
            dict: outcome data
        """
        url = f"{self.session.root_url}/epi/api/outcome/{outcome_id}/"
        return self.session.get(url).json()

    def create_outcome(self, data: dict) -> dict:
        """
        Create a new outcome

        Args:
            data (dict): fields to create on the outcome. Example keys:
                name (str): name of the outcome
                assessment (int): id of the associated assessment
                study_population (int): id of the associated study population
                system (str): system
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/outcome/"
        return self.session.post(url, data).json()

    def update_outcome(self, outcome_id: int, data: dict) -> dict:
        """
        Update an existing outcome

        Args:
            outcome_id (int): outcome ID

            data: fields to update in the outcome
                See "create_outcome" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/outcome/{outcome_id}/"
        return self.session.patch(url, data).json()

    def delete_outcome(self, outcome_id: int) -> Response:
        """
        Delete an outcome

        Args:
            outcome_id (int): outcome ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/outcome/{outcome_id}/"
        return self.session.delete(url)

    def get_result(self, result_id: int) -> dict:
        """
        Retrieves data for a single result

        Args:
            result_id (int): result ID

        Returns:
            dict: result data
        """
        url = f"{self.session.root_url}/epi/api/result/{result_id}/"
        return self.session.get(url).json()

    def create_result(self, data: dict) -> dict:
        """
        Create a new result

        Args:
            data (dict): fields to create on the result. Example keys:
                name (str): name of the result
                comments (str): comments on this result
                outcome (int): id of the associated outcome
                comparison_set (int): id of the associated comparison_set
                factors_applied (list[str]): list of factor names to apply
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/result/"
        return self.session.post(url, data).json()

    def update_result(self, result_id: int, data: dict) -> dict:
        """
        Update an existing result

        Args:
            result_id (int): result ID

            data: fields to update in the result
                See "create_result" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/result/{result_id}/"
        return self.session.patch(url, data).json()

    def delete_result(self, result_id: int) -> Response:
        """
        Delete a result

        Args:
            result_id (int): result ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/result/{result_id}/"
        return self.session.delete(url)

    def get_group_result(self, group_result_id: int) -> dict:
        """
        Retrieves data for a single group result

        Args:
            group_result_id (int): group result ID

        Returns:
            dict: group result data
        """
        url = f"{self.session.root_url}/epi/api/group-result/{group_result_id}/"
        return self.session.get(url).json()

    def create_group_result(self, data: dict) -> dict:
        """
        Create a new group result

        Args:
            data (dict): fields to create on the group result. Example keys:
                name (str): name of the result
                group (int): id of the associated group
                result (int): id of the associated result
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/group-result/"
        return self.session.post(url, data).json()

    def update_group_result(self, group_result_id: int, data: dict) -> dict:
        """
        Update an existing group result

        Args:
            group_result_id (int): group result ID

            data: fields to update in the group result
                See "create_group_result" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/group-result/{group_result_id}/"
        return self.session.patch(url, data).json()

    def delete_group_result(self, group_result_id: int) -> Response:
        """
        Delete a group result

        Args:
            group_result_id (int): group result ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/group-result/{group_result_id}/"
        return self.session.delete(url)

    def get_exposure(self, exposure_id: int) -> dict:
        """
        Retrieves data for a single exposure

        Args:
            exposure_id (int): exposure ID

        Returns:
            dict: exposure data
        """
        url = f"{self.session.root_url}/epi/api/exposure/{exposure_id}/"
        return self.session.get(url).json()

    def create_exposure(self, data: dict) -> dict:
        """
        Create a new exposure

        Args:
            data (dict): fields to create on the exposure. Example keys:
                name (str): name of the exposure
                study_population (int): id of the associated study population
                dtxsid (str): e.g. "DTXSID1020190",
                central_tendendices (list[dict]): list of CT's with keys like "estimate", "estimate_type", "lower_ci", etc.
                etc.

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epi/api/exposure/"
        return self.session.post(url, data).json()

    def update_exposure(self, exposure_id: int, data: dict) -> dict:
        """
        Update an existing exposure

        Args:
            exposure_id (int): exposure ID

            data: fields to update in the exposure
                See "create_exposure" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epi/api/exposure/{exposure_id}/"
        return self.session.patch(url, data).json()

    def delete_exposure(self, exposure_id: int) -> Response:
        """
        Delete an exposure

        Args:
            exposure_id (int): exposure ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epi/api/exposure/{exposure_id}/"
        return self.session.delete(url)



from urllib.parse import urlencode

from requests import Response

from .client import BaseClient


class AssessmentClient(BaseClient):
    """
    Client class for assessment requests.
    """

    def public(self) -> list[dict]:
        """
        Retrieves data pertaining to all public assessments

        Returns:
            list[dict]: Public assessment data
        """
        url = f"{self.session.root_url}/assessment/api/assessment/public/"
        return self.session.get(url).json()

    def team_member(self) -> list[dict]:
        """
        Retrieves assessments where use is a team member or project manager

        Returns:
            list[dict]: Public assessment data
        """
        url = f"{self.session.root_url}/assessment/api/assessment/team-member/"
        return self.session.get(url).json()

    def create_value(self, data: dict) -> dict:
        """
        Create an assessment value.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/assessment/api/value/"
        return self.session.post(url, data).json()

    def create_detail(self, data: dict) -> dict:
        """
        Create additional details for an assessment.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/assessment/api/detail/"
        return self.session.post(url, data).json()

    def all_values(self) -> list[dict]:
        """
        Get a list of all assessment values in HAWC (admin only).

        Returns:
            list[dict]: Assessment values data
        """
        url = f"{self.session.root_url}/admin/api/reports/values/"
        return self.session.get(url).json()

    def create(self, data: dict) -> dict:
        """
        Create an assessment.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/assessment/api/assessment/"
        return self.session.post(url, data).json()

    def update(self, assessment_id: int, data: dict) -> dict:
        """
        Update an existing assessment

        Args:
            assessment_id (int): assessment ID

            data: fields to update in the assessment

        Returns:
            dict: The resulting object, if update was successful
        """
        url = f"{self.session.root_url}/assessment/api/assessment/{assessment_id}/"
        return self.session.patch(url, data).json()

    def delete(self, assessment_id: int) -> Response:
        """
        Delete a assessment

        Args:
            assessment_id (int): assessment ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/assessment/api/assessment/{assessment_id}/"
        return self.session.delete(url)

    def effect_tag_create(self, name: str, slug: str) -> dict:
        """Create an effect tag.

        Effect tags can be associated with endpoints our health outcomes.

        Args:
            name (str): Tag name
            slug (str): Tag slug

        Returns:
            dict: The created tag
        """
        url = f"{self.session.root_url}/assessment/api/effect-tag/"
        return self.session.post(url, {"name": name, "slug": slug}).json()

    def effect_tag_list(self, name: str = "") -> dict:
        """Return a paginated list of available tags.

        If a name is specified, list is filtered to tags with the specified name.

        Args:
            name (str, optional): A search term to filter tag listing

        Returns:
            dict: a list of search results
        """
        url = f"{self.session.root_url}/assessment/api/effect-tag/"
        if name:
            url += "?" + urlencode({"name": name})
        return self.session.get(url).json()



import pandas as pd
from requests import Response

from .client import BaseClient


class EpiV2Client(BaseClient):
    """
    Client class for epidemiology v2 requests.
    """

    def metadata(self) -> dict:
        """
        Retrieves field choices for all epiv2 models.

        Returns:
            dict: Model metadata
        """
        url = f"{self.session.root_url}/epidemiology/api/metadata/"
        return self.session.get(url).json()

    def data(self, assessment_id: int, retrieve_unpublished_data: bool = False) -> pd.DataFrame:
        """
        Retrieves flat epidemiology v2 data export for the given assessment.

        Args:
            assessment_id (int): Assessment ID
            retrieve_unpublished_data (bool): include unpublished data in returned DataFrame

        Returns:
            pd.DataFrame: Epidemiology data
        """
        url = f"{self.session.root_url}/epidemiology/api/assessment/{assessment_id}/export/"
        if retrieve_unpublished_data:
            url += "?unpublished=true"

        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def get_designs_for_assessment(self, assessment_id: int, page: int = 1) -> pd.DataFrame:
        """
        Retrieves all of the designs for the given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: design data
        """
        url = f"{self.session.root_url}/epidemiology/api/design/?assessment_id={assessment_id}&page={page}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def get_designs_for_study(
        self, assessment_id: int, study_id: int, page: int = 1
    ) -> pd.DataFrame:
        """
        Retrieves all of the designs for the given study.

        Args:
            assessment_id (int): Assessment ID
            study_id (int): Study ID

        Returns:
            pd.DataFrame: design data
        """
        url = f"{self.session.root_url}/epidemiology/api/design/?assessment_id={assessment_id}&study={study_id}&page={page}"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def get_design(self, design_id: int) -> dict:
        """
        Retrieves data for a single design

        Args:
            design_id (int): design ID

        Returns:
            dict: design data
        """
        url = f"{self.session.root_url}/epidemiology/api/design/{design_id}/"
        return self.session.get(url).json()

    def create_design(self, data: dict) -> dict:
        """
        Create a new design

        Args:
            data (dict): fields to create on the design. Keys:
                study_id (int): id of the associated study
                study_design (str): design (predefined choices: 'Cohort', 'Case-control', etc.)
                age_profile (list[str]): age profile (predefined choices: 'Adults', 'Pregnant women', etc.)
                source (str): source (predefined choices: 'General population', Occupational', etc.)
                sex (str): sex (predefined choices: 'Female', 'Not reported', etc.)
                countries (list[str]): list of country codes to associate with the design
                region (str): other geographic information (free-text)
                summary (str): description of the study population (free-text)
                study_name (str): study name assigned by the authors (free-text)
                age_description (str): population age details (free-text)
                race (str): population race/ethnicity (free-text)
                participant_n (int): total # of participants enrolled in the study
                years_enrolled (str): years of enrollment (free-text)
                years_followup (str): years/length of follow-up (free-text)
                criteria (str): inclusion/exclusion criteria (free-text)
                susceptibility (str): notes on if study presents information for potentially susceptible/vulnerable populations or sub-populations (free-text)
                comments (str): comments (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/design/"
        return self.session.post(url, data).json()

    def update_design(self, design_id: int, data: dict) -> dict:
        """
        Update an existing design

        Args:
            design_id (int): design ID
            data: fields to update in the design
                See "create_design" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/design/{design_id}/"
        return self.session.patch(url, data).json()

    def delete_design(self, design_id: int) -> Response:
        """
        Delete a design

        Args:
            design_id (int): design ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/design/{design_id}/"
        return self.session.delete(url)

    def get_chemical(self, chemical_id: int) -> dict:
        """
        Retrieves data for a single chemical

        Args:
            chemical_id (int): chemical ID

        Returns:
            dict: chemical data
        """
        url = f"{self.session.root_url}/epidemiology/api/chemical/{chemical_id}/"
        return self.session.get(url).json()

    def create_chemical(self, data: dict) -> dict:
        """
        Create a new chemical

        Args:
            data (dict): fields to create on the chemical. Keys:
                design (int): id of the associated design
                name (str): name of the chemical
                dsstox_id (str): DSSTox substance identifier

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/chemical/"
        return self.session.post(url, data).json()

    def update_chemical(self, chemical_id: int, data: dict) -> dict:
        """
        Update an existing chemical

        Args:
            chemical_id (int): chemical ID
            data: fields to update in the chemical
                See "create_chemical" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/chemical/{chemical_id}/"
        return self.session.patch(url, data).json()

    def delete_chemical(self, chemical_id: int) -> Response:
        """
        Delete a chemical

        Args:
            chemical_id (int): chemical ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/chemical/{chemical_id}/"
        return self.session.delete(url)

    def get_exposure(self, exposure_id: int) -> dict:
        """
        Retrieves data for a single exposure

        Args:
            exposure_id (int): exposure ID

        Returns:
            dict: exposure data
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure/{exposure_id}/"
        return self.session.get(url).json()

    def create_exposure(self, data: dict) -> dict:
        """
        Create a new exposure

        Args:
            data (dict): fields to create on the exposure. Keys:
                design (int): id of the associated design
                name (str): unique name for the exposure (free-text)
                measurement_type (list[str]): measurement types (predefined choices: 'Biomonitoring', 'Air', etc.)
                measurement_timing (str): age, other timing, cross-sectional, etc. (free-text)
                exposure_route (str): most appropriate route (predefined choices: 'Inhalation', 'Oral', etc.)
                measurement_method (str): method used to measure exposure (free-text)
                comments (str): general comments (free-text)
                biomonitoring_matrix (str): if measurement_type does not includes 'Biomonitoring', use empty string (free-text)
                biomonitoring_source (str): if measurement_type does not include 'Biomonitoring', use empty string (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure/"
        return self.session.post(url, data).json()

    def update_exposure(self, exposure_id: int, data: dict) -> dict:
        """
        Update an existing exposure

        Args:
            exposure_id (int): exposure ID
            data: fields to update in the exposure
                See "create_exposure" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/exposure/{exposure_id}/"
        return self.session.patch(url, data).json()

    def delete_exposure(self, exposure_id: int) -> Response:
        """
        Delete an exposure

        Args:
            exposure_id (int): exposure ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure/{exposure_id}/"
        return self.session.delete(url)

    def get_exposure_level(self, exposure_level_id: int) -> dict:
        """
        Retrieves data for a single exposure level

        Args:
            exposure_level_id (int): exposure level ID

        Returns:
            dict: exposure level data
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure-level/{exposure_level_id}/"
        return self.session.get(url).json()

    def create_exposure_level(self, data: dict) -> dict:
        """
        Create a new exposure level

        Args:
            data (dict): fields to create on the exposure level. Keys:
                design (int): id of the associated design
                chemical_id (int): id of the associated chemical
                exposure_measurement_id (int): id of the associated exposure
                name (str): unique name for the exposure level (free-text)
                sub_population (str): subgroup within the study population (free-text)
                median (float)
                mean (float)
                variance (float)
                variance_type (str): measure of variation reported (predefined choices: 'N/A', 'SD', etc.)
                units (str): (free-text)
                ci_lcl (str): lower interval (free-text, non-numeric values should be limited to '<', 'LOD', etc.)
                percentile_25 (str): 25th percentile (free-text, non-numeric values should be limited to '<', 'LOD', etc.)
                percentile_75 (str): 75th percentile (free-text, non-numeric values should be limited to '<', 'LOD', etc.)
                ci_ucl (str): upper interval (free-text, non-numeric values should be limited to '<', 'LOD', etc.)
                ci_type (str): lower/upper interval type (predefined choices: 'Rng', 'P90', 'Oth', etc.)
                negligible_exposure (str): % of population without measurable exposure (free-text)
                data_location (str): e.g. table number (free-text)
                comments (str): general comments (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure-level/"
        return self.session.post(url, data).json()

    def update_exposure_level(self, exposure_level_id: int, data: dict) -> dict:
        """
        Update an existing exposure level

        Args:
            exposure_level_id (int): exposure level ID
            data: fields to update in the exposure level
                See "create_exposure_level" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/exposure-level/{exposure_level_id}/"
        return self.session.patch(url, data).json()

    def delete_exposure_level(self, exposure_level_id: int) -> Response:
        """
        Delete an exposure level

        Args:
            exposure_level_id (int): exposure level ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/exposure-level/{exposure_level_id}/"
        return self.session.delete(url)

    def get_outcome(self, outcome_id: int) -> dict:
        """
        Retrieves data for a single outcome

        Args:
            outcome_id (int): outcome ID

        Returns:
            dict: outcome data
        """
        url = f"{self.session.root_url}/epidemiology/api/outcome/{outcome_id}/"
        return self.session.get(url).json()

    def create_outcome(self, data: dict) -> dict:
        """
        Create a new outcome

        Args:
            data (dict): fields to create on the outcome. Keys:
                design (int): id of the associated design
                system (str): most relevant system (predefined choices: 'CA'/'Cancer', 'DE'/'Dermal', etc.)
                effect (str): health effect of interest (free-text)
                effect_detail (str): optional additional effect specification (free-text)
                endpoint (str): unique name for the specific endpoint/outcome being measure (free-text)
                comments (str): comments (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/outcome/"
        return self.session.post(url, data).json()

    def update_outcome(self, outcome_id: int, data: dict) -> dict:
        """
        Update an existing outcome

        Args:
            outcome_id (int): outcome ID
            data: fields to update in the outcome
                See "create_outcome" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/outcome/{outcome_id}/"
        return self.session.patch(url, data).json()

    def delete_outcome(self, outcome_id: int) -> Response:
        """
        Delete an outcome

        Args:
            outcome_id (int): outcome ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/outcome/{outcome_id}/"
        return self.session.delete(url)

    def get_adjustment_factor(self, adjustment_factor_id: int) -> dict:
        """
        Retrieves data for a single adjustment factor

        Args:
            adjustment_factor_id (int): adjustment factor ID

        Returns:
            dict: adjustment factor data
        """
        url = f"{self.session.root_url}/epidemiology/api/adjustment-factor/{adjustment_factor_id}/"
        return self.session.get(url).json()

    def create_adjustment_factor(self, data: dict) -> dict:
        """
        Create a new adjustment factor

        Args:
            data (dict): fields to create on the adjustment factor. Keys:
                design (int): id of the associated design
                name (str): unique name for the adjustment factor (free-text)
                description (str): comma-separated list of covariates, using uniform language across studies if possible (free-text)
                comments (str): general comments (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/adjustment-factor/"
        return self.session.post(url, data).json()

    def update_adjustment_factor(self, adjustment_factor_id: int, data: dict) -> dict:
        """
        Update an existing adjustment factor

        Args:
            adjustment_factor_id (int): adjustment factor ID
            data: fields to update in the adjustment factor
                See "create_adjustment_factor" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/adjustment-factor/{adjustment_factor_id}/"
        return self.session.patch(url, data).json()

    def delete_adjustment_factor(self, adjustment_factor_id: int) -> Response:
        """
        Delete an adjustment factor

        Args:
            adjustment_factor_id (int): adjustment factor ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/adjustment-factor/{adjustment_factor_id}/"
        return self.session.delete(url)

    def get_data_extraction(self, data_extraction_id: int) -> dict:
        """
        Retrieves data for a single data extraction

        Args:
            data_extraction_id (int): data extraction ID

        Returns:
            dict: data extraction data
        """
        url = f"{self.session.root_url}/epidemiology/api/data-extraction/{data_extraction_id}/"
        return self.session.get(url).json()

    def create_data_extraction(self, data: dict) -> dict:
        """
        Create a new data extraction

        Args:
            data (dict): fields to create on the data extraction. Keys:
                design (int): id of the associated design
                outcome_id (int): id of the associated outcome
                exposure_level_id (int): id of the associated exposure level
                factors_id (int): id of the associated adjustment factors
                sub_population (str): subgroup within the study population (free-text)
                outcome_measurement_timing (str): age or other timing, "cross-sectional", etc. (free-text)
                effect_estimate_type (str): type of effect estimate (predefined choices: 'Odds Ratio'/'OR', 'Absolute Risk %'/'AR', etc.)
                effect_estimate (float): value of effect estimate
                ci_lcl (float): lower bound
                ci_ucl (float): upper bound
                ci_type (str): lower/upper bound type (predefined choices: 'P90', '5th/95h percentile', etc.)
                units (str): bound units (free-text)
                variance_type (str): type of variance estimate (predefined choices: 'SD', 'SE', 'NA', etc.)
                variance (float): value of variance estimate
                n (int): n
                p_value (str): p-value (free-text)
                significant (str): was the data extraction statistically significant? (predefined choices: 'NR', 'Yes', etc.)
                group (str): results group for linking sets of results (free-text)
                exposure_rank (int): value for ordering linked sets of results
                exposure_transform (str): exposure transform (predefined choices: 'log10', 'squared', 'other', etc.)
                outcome_transform (str): outcome transform (predefined choices: 'log10', 'squared', 'other', etc.)
                confidence (str): overall confidence rating for the endpoint being extracted (free-text)
                adverse_direction (str): direction of effect that would be adverse if observed (predefined choices: 'unknown', 'up', 'any', etc.)
                data_location (str): e.g. table number (free-text)
                effect_description (str): description of the effect estimate with units, including comparison being made. (free-text)
                statistical_method (str): brief description of the statistical analysis method (free-text)
                comments (str): general comments (free-text)

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/epidemiology/api/data-extraction/"
        return self.session.post(url, data).json()

    def update_data_extraction(self, data_extraction_id: int, data: dict) -> dict:
        """
        Update an existing data extraction

        Args:
            data_extraction_id (int): data extraction ID
            data: fields to update in the data extraction
                See "create_data_extraction" docstring for example fields.

        Returns:
            dict: The resulting object, if update was successful
        """

        url = f"{self.session.root_url}/epidemiology/api/data-extraction/{data_extraction_id}/"
        return self.session.patch(url, data).json()

    def delete_data_extraction(self, data_extraction_id: int) -> Response:
        """
        Delete an data extraction

        Args:
            data_extraction_id (int): data extraction ID

        Returns:
            Response: The response object.
        """
        url = f"{self.session.root_url}/epidemiology/api/data-extraction/{data_extraction_id}/"
        return self.session.delete(url)



from io import StringIO

import pandas as pd

from .session import HawcSession


class BaseClient:
    """
    Base client class.

    Initiates with a given HawcSession object.
    """

    def __init__(self, session: HawcSession):
        self.session = session

    def _csv_to_df(self, csv: str) -> pd.DataFrame:
        """
        Takes a CSV string and returns the pandas DataFrame representation of it.

        Args:
            csv (str): CSV string

        Returns:
            pd.DataFrame: DataFrame from CSV
        """
        csv_io = StringIO(csv)
        return pd.read_csv(csv_io)



from copy import deepcopy
from operator import itemgetter

import pandas as pd

from .client import BaseClient


class AnimalClient(BaseClient):
    """
    Client class for animal experiment requests.
    """

    def create_experiment(self, data: dict) -> dict:
        """
        Create a new experiment.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/ani/api/experiment/"
        return self.session.post(url, data).json()

    def create_animal_group(self, data: dict) -> dict:
        """
        Create a new animal-group and dosing regime.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/ani/api/animal-group/"
        return self.session.post(url, data).json()

    def create_endpoint(self, data: dict) -> dict:
        """
        Create a new endpoint.

        Args:
            data (dict): required metadata for creation

        Returns:
            dict: The resulting object, if create was successful
        """
        url = f"{self.session.root_url}/ani/api/endpoint/"
        return self.session.post(url, data).json()

    def data(self, assessment_id: int, include_unpublished: bool = False) -> pd.DataFrame:
        """
        Retrieves a complete export of animal bioassay data for a given assessment.

        Args:
            assessment_id (int): Assessment ID
            include_unpublished (bool, optional): If True, includes data from unpublished studies. Defaults to False.

        Returns:
            pd.DataFrame: Complete bioassay export
        """
        url = f"{self.session.root_url}/ani/api/assessment/{assessment_id}/full-export/"
        if include_unpublished:
            url += "?unpublished=true"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def data_summary(self, assessment_id: int) -> pd.DataFrame:
        """
        Retrieves an endpoint summary of animal bioassay data for a given assessment.

        Args:
            assessment_id (int): Assessment ID

        Returns:
            pd.DataFrame: Endpoint bioassay summary
        """
        url = f"{self.session.root_url}/ani/api/assessment/{assessment_id}/endpoint-export/"
        response_json = self.session.get(url).json()
        return pd.DataFrame(response_json)

    def _invert_endpoints(self, endpoints: list[dict]) -> list[dict]:
        studies = {}
        for endpoint in deepcopy(endpoints):
            study = endpoint["animal_group"]["experiment"]["study"]
            if study["id"] not in studies:
                study["experiments"] = {}
                studies[study["id"]] = study
            matched_study = studies[study["id"]]

            experiment = endpoint["animal_group"]["experiment"]
            if experiment["id"] not in matched_study["experiments"]:
                experiment["animal_groups"] = {}
                matched_study["experiments"][experiment["id"]] = experiment
            matched_experiment = matched_study["experiments"][experiment["id"]]

            animal_group = endpoint["animal_group"]
            if animal_group["id"] not in matched_experiment["animal_groups"]:
                animal_group["endpoints"] = []
                matched_experiment["animal_groups"][animal_group["id"]] = animal_group
            matched_animal_group = matched_experiment["animal_groups"][animal_group["id"]]

            matched_animal_group["endpoints"].append(endpoint)

        # cleanup
        studies = sorted(list(studies.values()), key=itemgetter("id"))
        for study in studies:
            study["experiments"] = sorted(list(study["experiments"].values()), key=itemgetter("id"))
            for experiment in study["experiments"]:
                experiment.pop("study")
                experiment["animal_groups"] = sorted(
                    list(experiment["animal_groups"].values()), key=itemgetter("id")
                )
                for animal_group in experiment["animal_groups"]:
                    animal_group.pop("experiment")
                    for endpoint in animal_group["endpoints"]:
                        endpoint.pop("animal_group")
                    animal_group["endpoints"] = sorted(
                        animal_group["endpoints"], key=itemgetter("id")
                    )

        return studies

    def endpoints(self, assessment_id: int, invert: bool = False) -> list[dict]:
        """
        Retrieves all bioassay endpoints for a given assessment.

        Args:
            assessment_id (int): Assessment ID
            invert (bool, default False): Return a list of endpoints if False (default). This is
                akin to data bottom up. If True, returns a list of studies instead, with related
                data all the way down to endpoints nested within the list of studies. This is
                akin to top-down.

        Returns:
            list[dict]: A list of endpoints and related studies with each (or a list of studies with
                a related endpoints if inverted)
        """
        payload = {"assessment_id": assessment_id}
        url = f"{self.session.root_url}/ani/api/endpoint/"
        generator = self.session.iter_pages(url, payload)
        data = [res for results in generator for res in results]
        if invert:
            data = self._invert_endpoints(data)
        return data

    def bmds_endpoints(self, assessment_id: int, unpublished: bool = False) -> pd.DataFrame:
        url = f"{self.session.root_url}/ani/api/assessment/{assessment_id}/bmds-export/"
        params = {"unpublished": unpublished}
        response_json = self.session.get(url, params=params).json()
        return pd.DataFrame(response_json)

    def metadata(self) -> dict:
        """
        Retrieves field choices for all animal models.

        Returns:
            dict: Model metadata
        """
        url = f"{self.session.root_url}/ani/api/metadata/"
        return self.session.get(url).json()



import time
from io import BytesIO
from pathlib import Path

from playwright._impl._api_structures import SetCookieParam
from playwright.async_api import Page, TimeoutError, expect
from playwright.async_api._context_manager import PlaywrightContextManager as pcm

from .client import BaseClient
from .exceptions import ContentUnavailable, HawcClientException


async def remove_dj_toolbar(page: Page):
    if await page.evaluate("document.querySelector('#djDebug')"):
        await page.evaluate("document.querySelector('#djDebug').remove()")


async def get_tableau_image(page) -> BytesIO:
    download_button = page.frame_locator("iframe").locator(
        'div[role="button"]:has-text("Download")'
    )
    download_confirm = page.frame_locator("iframe").locator('button:has-text("Image")')
    await download_button.click()
    async with page.expect_download() as download_info:
        await download_confirm.click()
    download = await download_info.value
    path = await download.path()
    return BytesIO(path.read_bytes())


async def get_plotly_image(page) -> BytesIO:
    download_confirm = page.locator(".js-plotly-plot .modebar-btn").first
    async with page.expect_download() as download_info:
        await download_confirm.click()
    download = await download_info.value
    path = await download.path()
    return BytesIO(path.read_bytes())


async def get_static_image(page) -> BytesIO:
    b = await page.locator("#visual-image img").screenshot(type="png")
    return BytesIO(b)


async def get_svg_image(page, timeout_ms) -> BytesIO:
    error_alert = page.get_by_test_id("error")
    download_button = page.locator("button", has=page.locator("i.fa-download"))
    download_confirm = page.locator("text=Download as a PNG")

    try:
        await expect(download_button.or_(error_alert)).to_be_visible(timeout=timeout_ms)
    except AssertionError as err:
        raise TimeoutError(f"Timeout - exceeded {timeout_ms}ms") from err

    if await error_alert.is_visible():
        error_text = await error_alert.inner_text()
        raise ContentUnavailable(400, error_text)

    # wait 1 second; the tagtree has startup animations
    time.sleep(1)

    await download_button.click()
    async with page.expect_download() as download_info:
        await download_confirm.click()
    download = await download_info.value
    path = await download.path()
    return BytesIO(path.read_bytes())


async def fetch_png(page: Page, timeout_ms: int) -> BytesIO:
    """Helper method to download a PNG from a visualization page

    Args:
        page (Page): a page instance
        timeout_ms (int): timeout, in milliseconds

    Returns:
        BytesIO: The PNG image, in bytes
    """
    await page.wait_for_load_state("load")
    await expect(page.locator(".is-loading")).to_have_count(0)
    await remove_dj_toolbar(page)

    viz_type = await page.evaluate(
        "document.querySelector('meta[name=hawc-viz-type]').dataset.vizType"
    )
    match viz_type:
        case (
            "data pivot"
            | "animal bioassay endpoint aggregation"
            | "animal bioassay endpoint crossview"
            | "risk of bias heatmap"
            | "study evaluation heatmap"
            | "risk of bias barchart"
            | "study evaluation barchart"
            | "literature tagtree"
            | "exploratory heatmap"
        ):
            return await get_svg_image(page, timeout_ms)
        case "embedded external website":
            return await get_tableau_image(page)
        case "plotly":
            return await get_plotly_image(page)
        case "static image":
            return await get_static_image(page)
        case _:
            raise ValueError(f"Unknown visual type: {viz_type}")


PathLike = Path | str | None


def write_to_file(data: BytesIO, path: PathLike) -> None:
    """Write to a file, given a path-like object"""
    if path is None:
        return
    if isinstance(path, str):
        path = Path(path)
    path.write_bytes(data.getvalue())


class InteractiveHawcClient:
    """
    A context manager for downloading assessment visuals.
    """

    def __init__(self, client: BaseClient, headless: bool = True, timeout: float = 60):
        """Start an interactive HAWC client to process content.

        Args:
            client (BaseClient): the current client; requires `login=True` to process content
                which requires authentication.
            headless (bool): Run in headless mode; defaults to True.
            timeout (float): Timeout, in seconds; defaults to 60
        """
        self.client = client
        self.headless = headless
        self.timeout = timeout * 1000  # convert from s to ms

    async def __aenter__(self):
        self.playwright = await pcm().start()
        browser = await self.playwright.chromium.launch(headless=self.headless)
        self.context = await browser.new_context()
        if self.timeout:
            self.context.set_default_timeout(self.timeout)
        self.page = await self.context.new_page()
        cookies = [
            SetCookieParam(name=k, value=v, url=self.client.session.root_url)
            for k, v in self.client.session._session.cookies.items()
        ]
        if not cookies:
            raise HawcClientException(
                403,
                "No cookies found on client session;\nwhen setting authorization token; set login to True",
            )
        await self.context.add_cookies(cookies)
        return self

    async def __aexit__(self, *args) -> None:
        await self.context.close()
        await self.playwright.stop()

    async def download_visual(self, id: int, fn: PathLike = None) -> BytesIO:
        """Download a PNG visualization given a visual ID

        Args:
            id (int): The visual ID
            fn (PathLike, optional): If a path or string is specified, the PNG is written to
                that location. If None (default), no data is written to a Path.

        Returns:
            BytesIO: the PNG representation of the visual, in bytes.
        """
        url = f"{self.client.session.root_url}/summary/visual/{id}/"
        # ensure response is OK before waiting
        response = await self.page.goto(url)
        if response and not response.ok:
            raise HawcClientException(response.status, response.status_text)
        self.page.set_default_timeout(self.timeout)
        data = await fetch_png(self.page, self.timeout)
        write_to_file(data, fn)
        return data

    async def download_data_pivot(self, id: int, fn: PathLike = None) -> BytesIO:
        """Download a PNG data pivot given a data pivot ID

        Args:
            id (int): The data pivot ID
            fn (PathLike, optional): If a path or string is specified, the a PNG is written to
                that location. If None (default), no data is written to a Path.

        Returns:
            BytesIO: the PNG representation of the data pivot, in bytes.
        """
        url = f"{self.client.session.root_url}/summary/data-pivot/{id}/"
        # ensure response is OK before waiting
        response = await self.page.goto(url)
        if response and not response.ok:
            raise HawcClientException(response.status, response.status_text)
        # attempt to fetch PNG
        self.page.set_default_timeout(self.timeout)
        data = await fetch_png(self.page, self.timeout)
        write_to_file(data, fn)
        return data



from .client import BaseClient


class VocabClient(BaseClient):
    """
    Client class for vocabulary requests.
    """

    def bulk_create(self, terms: list[dict]) -> list[dict]:
        """
        Bulk creates a list of terms.

        Args:
            terms (list[dict]): List of serialized terms.

        Returns:
            list[dict]: List of created, serialized terms.
        """
        url = f"{self.session.root_url}/vocab/api/term/bulk-create/"
        return self.session.post(url, terms).json()

    def bulk_update(self, terms: list[dict]) -> list[dict]:
        """
        Bulk updates a list of terms.

        Args:
            terms (list[dict]): List of serialized terms.

        Returns:
            list[dict]: List of updated, serialized terms.
        """
        url = f"{self.session.root_url}/vocab/api/term/bulk-update/"
        return self.session.patch(url, terms).json()

    def uids(self) -> list[tuple[int, int]]:
        """
        Get all term ids and uids.

        Returns:
            list[tuple[int,int]]: List of id, uid tuples for all terms.
        """
        url = f"{self.session.root_url}/vocab/api/term/uids/"
        return self.session.get(url).json()
